from . import madoka

VERSION = (0, 7, 1)
__version__ = '0.7.1'
__all__ = ['Sketch', 'CroquisUint8', 'CroquisUint16', 'CroquisUint32',
           'CroquisUint64', 'CroquisFloat', 'CroquisDouble']

Sketch = madoka.Sketch
CroquisUint8 = madoka.CroquisUint8
CroquisUint16 = madoka.CroquisUint16
CroquisUint32 = madoka.CroquisUint32
CroquisUint64 = madoka.CroquisUint64
CroquisFloat = madoka.CroquisFloat
CroquisDouble = madoka.CroquisDouble
